import * as ts from 'typescript';
import { Set } from './base';
export interface TranspilerOptions {
    /**
     * Fail on the first error, do not collect multiple. Allows easier debugging as stack traces lead
     * directly to the offending line.
     */
    failFast?: boolean;
    /** Whether to generate 'library a.b.c;' names from relative file paths. */
    generateLibraryName?: boolean;
    /**
     * A base path to relativize absolute file paths against. This is useful for library name
     * generation (see above) and nicer file names in error messages.
     */
    basePath?: string;
    /**
     * Use dart:html instead of the raw JavaScript DOM when generated Dart code.
     */
    useHtml?: boolean;
    /**
     * Enforce conventions of public/private keyword and underscore prefix
     */
    enforceUnderscoreConventions?: boolean;
    /**
     * Sets a root path to look for typings used by the facade converter.
     */
    typingsRoot?: string;
    /**
     * Experimental JS Interop specific option to promote properties with function
     * types to methods instead of properties with a function type. This the makes
     * the Dart code more readable at the cost of disallowing setting the value of
     * the property.
     * Example JS library that benifits from this option:
     * https://github.com/DefinitelyTyped/DefinitelyTyped/blob/master/chartjs/chart.d.ts
     */
    promoteFunctionLikeMembers?: boolean;
}
export declare const COMPILER_OPTIONS: ts.CompilerOptions;
/**
 * Context to ouput code into.
 */
export declare enum OutputContext {
    Import = 0,
    Header = 1,
    Default = 2,
}
export declare class Transpiler {
    private options;
    private outputs;
    private outputStack;
    private currentFile;
    importsEmitted: Set;
    private lastCommentIdx;
    private errors;
    private transpilers;
    private declarationTranspiler;
    private fc;
    private nameRewriter;
    constructor(options?: TranspilerOptions);
    /**
     * Transpiles the given files to Dart.
     * @param fileNames The input files.
     * @param destination Location to write files to. Creates files next to their sources if absent.
     */
    transpile(fileNames: string[], destination?: string): void;
    translateProgram(program: ts.Program): {
        [path: string]: string;
    };
    private getCompilerOptions();
    private createCompilerHost();
    getOutputPath(filePath: string, destinationRoot: string): string;
    pushContext(context: OutputContext): void;
    popContext(): void;
    private translate(sourceFile);
    private formatCode(code, context);
    private checkForErrors(program);
    /**
     * Returns `filePath`, relativized to the program's `basePath`.
     * @param filePath Optional path to relativize, defaults to the current file's path.
     */
    getRelativeFileName(filePath?: string): string;
    private readonly currentOutput;
    emit(s: string): void;
    emitNoSpace(s: string): void;
    enterCodeComment(): void;
    exitCodeComment(): void;
    reportError(n: ts.Node, message: string): void;
    visit(node: ts.Node): void;
    private normalizeSlashes(path);
    private translateComment(comment);
}
export declare function getModuleResolver(compilerHost: ts.CompilerHost): (moduleNames: string[], containingFile: string) => ts.ResolvedModule[];
