"use strict";
var ts = require('typescript');
var main_1 = require('./main');
function ident(n) {
    if (n.kind === ts.SyntaxKind.Identifier)
        return n.text;
    if (n.kind === ts.SyntaxKind.QualifiedName) {
        var qname = n;
        var leftName = ident(qname.left);
        if (leftName)
            return leftName + '.' + ident(qname.right);
    }
    return null;
}
exports.ident = ident;
function isFunctionTypedefLikeInterface(ifDecl) {
    return ifDecl.members && ifDecl.members.length === 1 &&
        ifDecl.members[0].kind === ts.SyntaxKind.CallSignature;
}
exports.isFunctionTypedefLikeInterface = isFunctionTypedefLikeInterface;
function getDeclaration(type) {
    var symbol = type.getSymbol();
    if (!symbol)
        return null;
    if (symbol.valueDeclaration)
        return symbol.valueDeclaration;
    return symbol.declarations && symbol.declarations.length > 0 ? symbol.declarations[0] : null;
}
exports.getDeclaration = getDeclaration;
function isExtendsClause(heritageClause) {
    return heritageClause.token === ts.SyntaxKind.ExtendsKeyword &&
        heritageClause.parent.kind !== ts.SyntaxKind.InterfaceDeclaration;
}
exports.isExtendsClause = isExtendsClause;
function isConstructor(n) {
    return n.kind === ts.SyntaxKind.Constructor || n.kind === ts.SyntaxKind.ConstructSignature;
}
exports.isConstructor = isConstructor;
function isStatic(n) {
    var hasStatic = false;
    ts.forEachChild(n, function (child) {
        if (child.kind === ts.SyntaxKind.StaticKeyword) {
            hasStatic = true;
        }
    });
    return hasStatic;
}
exports.isStatic = isStatic;
function isCallableType(type, tc) {
    if (isFunctionType(type, tc))
        return true;
    if (type.kind === ts.SyntaxKind.TypeReference) {
        if (tc.getSignaturesOfType(tc.getTypeAtLocation(type), ts.SignatureKind.Call).length > 0)
            return true;
    }
    return false;
}
exports.isCallableType = isCallableType;
function isFunctionType(type, tc) {
    var kind = type.kind;
    if (kind === ts.SyntaxKind.FunctionType)
        return true;
    if (kind === ts.SyntaxKind.TypeReference) {
        var t = tc.getTypeAtLocation(type);
        if (t.symbol && t.symbol.flags & ts.SymbolFlags.Function)
            return true;
    }
    if (kind === ts.SyntaxKind.UnionType) {
        var types = type.types;
        for (var i = 0; i < types.length; ++i) {
            if (!isFunctionType(types[i], tc)) {
                return false;
            }
        }
        return true;
    }
    // Warning: if the kind is a reference type and the reference is to an
    // interface that only has a call member we will not return that it is a
    // function type.
    if (kind === ts.SyntaxKind.TypeLiteral) {
        var members = type.members;
        for (var i = 0; i < members.length; ++i) {
            if (members[i].kind !== ts.SyntaxKind.CallSignature) {
                return false;
            }
        }
        return true;
    }
    return false;
}
exports.isFunctionType = isFunctionType;
function isTypeNode(node) {
    switch (node.kind) {
        case ts.SyntaxKind.UnionType:
        case ts.SyntaxKind.TypeReference:
        case ts.SyntaxKind.TypeLiteral:
        case ts.SyntaxKind.LastTypeNode:
        case ts.SyntaxKind.ArrayType:
        case ts.SyntaxKind.TypePredicate:
        case ts.SyntaxKind.TypeQuery:
        case ts.SyntaxKind.TupleType:
        case ts.SyntaxKind.NumberKeyword:
        case ts.SyntaxKind.StringKeyword:
        case ts.SyntaxKind.VoidKeyword:
        case ts.SyntaxKind.BooleanKeyword:
        case ts.SyntaxKind.AnyKeyword:
        case ts.SyntaxKind.FunctionType:
            return true;
        default:
            return false;
    }
}
exports.isTypeNode = isTypeNode;
function isCallable(decl) {
    var members = decl.members;
    return members.some(function (member) { return member.kind === ts.SyntaxKind.CallSignature; });
}
exports.isCallable = isCallable;
function copyLocation(src, dest) {
    dest.pos = src.pos;
    dest.end = src.end;
}
exports.copyLocation = copyLocation;
// Polyfill for ES6 Array.find.
function arrayFindPolyfill(nodeArray, predicate) {
    for (var i = 0; i < nodeArray.length; ++i) {
        if (predicate(nodeArray[i]))
            return nodeArray[i];
    }
    return null;
}
exports.arrayFindPolyfill = arrayFindPolyfill;
function getAncestor(n, kind) {
    for (var parent_1 = n; parent_1; parent_1 = parent_1.parent) {
        if (parent_1.kind === kind)
            return parent_1;
    }
    return null;
}
exports.getAncestor = getAncestor;
function getEnclosingClass(n) {
    for (var parent_2 = n.parent; parent_2; parent_2 = parent_2.parent) {
        if (parent_2.kind === ts.SyntaxKind.ClassDeclaration ||
            parent_2.kind === ts.SyntaxKind.InterfaceDeclaration) {
            return parent_2;
        }
    }
    return null;
}
exports.getEnclosingClass = getEnclosingClass;
function isConstCall(node) {
    return node && ident(node.expression) === 'CONST_EXPR';
}
exports.isConstCall = isConstCall;
function isInsideConstExpr(node) {
    return isConstCall(getAncestor(node, ts.SyntaxKind.CallExpression));
}
exports.isInsideConstExpr = isInsideConstExpr;
var TranspilerBase = (function () {
    function TranspilerBase(transpiler) {
        this.transpiler = transpiler;
        this.idCounter = 0;
    }
    TranspilerBase.prototype.visit = function (n) { this.transpiler.visit(n); };
    TranspilerBase.prototype.pushContext = function (context) { this.transpiler.pushContext(context); };
    TranspilerBase.prototype.popContext = function () { this.transpiler.popContext(); };
    TranspilerBase.prototype.emit = function (s) { this.transpiler.emit(s); };
    TranspilerBase.prototype.emitNoSpace = function (s) { this.transpiler.emitNoSpace(s); };
    TranspilerBase.prototype.enterCodeComment = function () { return this.transpiler.enterCodeComment(); };
    TranspilerBase.prototype.exitCodeComment = function () { return this.transpiler.exitCodeComment(); };
    TranspilerBase.prototype.emitImport = function (toEmit) {
        if (!this.transpiler.importsEmitted[toEmit]) {
            this.pushContext(main_1.OutputContext.Import);
            this.emit("import \"" + toEmit + "\";");
            this.transpiler.importsEmitted[toEmit] = true;
            this.popContext();
        }
    };
    TranspilerBase.prototype.reportError = function (n, message) { this.transpiler.reportError(n, message); };
    TranspilerBase.prototype.visitNode = function (n) { throw new Error('not implemented'); };
    TranspilerBase.prototype.visitEach = function (nodes) {
        var _this = this;
        nodes.forEach(function (n) { return _this.visit(n); });
    };
    TranspilerBase.prototype.visitEachIfPresent = function (nodes) {
        if (nodes)
            this.visitEach(nodes);
    };
    TranspilerBase.prototype.visitList = function (nodes, separator) {
        if (separator === void 0) { separator = ','; }
        for (var i = 0; i < nodes.length; i++) {
            this.visit(nodes[i]);
            if (i < nodes.length - 1)
                this.emit(separator);
        }
    };
    TranspilerBase.prototype.uniqueId = function (name) {
        var id = this.idCounter++;
        return "_" + name + "$$js_facade_gen$" + id;
    };
    TranspilerBase.prototype.assert = function (c, condition, reason) {
        if (!condition) {
            this.reportError(c, reason);
            throw new Error(reason);
        }
    };
    TranspilerBase.prototype.getAncestor = function (n, kind) {
        for (var parent_3 = n; parent_3; parent_3 = parent_3.parent) {
            if (parent_3.kind === kind)
                return parent_3;
        }
        return null;
    };
    TranspilerBase.prototype.hasAncestor = function (n, kind) { return !!getAncestor(n, kind); };
    TranspilerBase.prototype.hasAnnotation = function (decorators, name) {
        if (!decorators)
            return false;
        return decorators.some(function (d) {
            var decName = ident(d.expression);
            if (decName === name)
                return true;
            if (d.expression.kind !== ts.SyntaxKind.CallExpression)
                return false;
            var callExpr = d.expression;
            decName = ident(callExpr.expression);
            return decName === name;
        });
    };
    TranspilerBase.prototype.hasFlag = function (n, flag) {
        return n && (n.flags & flag) !== 0 || false;
    };
    TranspilerBase.prototype.getRelativeFileName = function (fileName) {
        return this.transpiler.getRelativeFileName(fileName);
    };
    TranspilerBase.prototype.maybeVisitTypeArguments = function (n) {
        if (n.typeArguments) {
            // If it's a single type argument `<void>`, ignore it and emit nothing.
            // This is particularly useful for `Promise<void>`, see
            // https://github.com/dart-lang/sdk/issues/2231#issuecomment-108313639
            if (n.typeArguments.length === 1 && n.typeArguments[0].kind === ts.SyntaxKind.VoidKeyword) {
                return;
            }
            this.emitNoSpace('<');
            this.visitList(n.typeArguments);
            this.emit('>');
        }
    };
    TranspilerBase.prototype.visitParameters = function (parameters) {
        this.emit('(');
        var firstInitParamIdx = 0;
        for (; firstInitParamIdx < parameters.length; firstInitParamIdx++) {
            // ObjectBindingPatterns are handled within the parameter visit.
            var isOpt = parameters[firstInitParamIdx].initializer ||
                parameters[firstInitParamIdx].questionToken ||
                parameters[firstInitParamIdx].dotDotDotToken;
            if (isOpt && parameters[firstInitParamIdx].name.kind !== ts.SyntaxKind.ObjectBindingPattern) {
                break;
            }
        }
        if (firstInitParamIdx !== 0) {
            var requiredParams = parameters.slice(0, firstInitParamIdx);
            this.visitList(requiredParams);
        }
        if (firstInitParamIdx !== parameters.length) {
            if (firstInitParamIdx !== 0)
                this.emit(',');
            var positionalOptional = parameters.slice(firstInitParamIdx, parameters.length);
            this.emit('[');
            this.visitList(positionalOptional);
            this.emit(']');
        }
        this.emit(')');
    };
    return TranspilerBase;
}());
exports.TranspilerBase = TranspilerBase;

//# sourceMappingURL=base.js.map
