"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var ts = require('typescript');
var base = require('./base');
var merge_1 = require('./merge');
function isFunctionLikeProperty(decl, tc) {
    if (!decl.type)
        return false;
    var name = base.ident(decl.name);
    if (name.match(/^on[A-Z]/))
        return false;
    return base.isFunctionType(decl.type, tc);
}
exports.isFunctionLikeProperty = isFunctionLikeProperty;
var DeclarationTranspiler = (function (_super) {
    __extends(DeclarationTranspiler, _super);
    function DeclarationTranspiler(tr, fc, enforceUnderscoreConventions, promoteFunctionLikeMembers) {
        _super.call(this, tr);
        this.fc = fc;
        this.enforceUnderscoreConventions = enforceUnderscoreConventions;
        this.promoteFunctionLikeMembers = promoteFunctionLikeMembers;
        this.moduleStack = [];
        this.extendsClass = false;
    }
    DeclarationTranspiler.prototype.setTypeChecker = function (tc) { this.tc = tc; };
    DeclarationTranspiler.prototype.setFacadeConverter = function (fc) { this.fc = fc; };
    DeclarationTranspiler.prototype.getJsPath = function (node) {
        var path = [].concat(this.moduleStack);
        var classDecl = base.getEnclosingClass(node);
        if (classDecl) {
            path.push(classDecl.name.text);
        }
        switch (node.kind) {
            case ts.SyntaxKind.ModuleDeclaration:
                break;
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.InterfaceDeclaration:
                path.push(node.name.text);
                break;
            case ts.SyntaxKind.EnumDeclaration:
                path.push(node.name.text);
                break;
            case ts.SyntaxKind.PropertyDeclaration:
            case ts.SyntaxKind.VariableDeclaration:
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
            case ts.SyntaxKind.PropertySignature:
                var memberName = base.ident(node.name);
                if (!base.isStatic(node) && classDecl != null)
                    return memberName;
                path.push(memberName);
                break;
            default:
                throw 'Internal error. Unexpected node kind:' + node.kind;
        }
        if (path.length === 1) {
            // No need to specify the path if is simply the node name.
            return '';
        }
        return path.join('.');
    };
    DeclarationTranspiler.prototype.isAnonymousInterface = function (node) {
        if (node.kind !== ts.SyntaxKind.InterfaceDeclaration)
            return false;
        // This is a bit of a hack but for the purposes of Dart codegen,
        // interfaces with static members or constructors have a known class name
        // at least for the purposes of resolving static members.
        // Example case that triggers this case:
        // interface Foo {
        //   bar();
        // }
        // declare let Foo: {
        //   new(): Foo,
        //   SOME_STATIC : number;
        // }
        return node.members.every(function (m) {
            return m.kind !== ts.SyntaxKind.Constructor && !base.isStatic(m);
        });
    };
    DeclarationTranspiler.prototype.maybeEmitJsAnnotation = function (node) {
        if (this.isAnonymousInterface(node)) {
            this.emit('@anonymous');
            this.emit('@JS()');
            return;
        }
        var name = this.getJsPath(node);
        this.emit('@JS(');
        if (name.length > 0) {
            this.emit('"' + name + '"');
        }
        this.emit(')');
    };
    /**
     * Emit fake constructors to placate the Dart Analyzer for JS Interop classes.
     */
    DeclarationTranspiler.prototype.maybeEmitFakeConstructors = function (decl) {
        if (decl.kind === ts.SyntaxKind.ClassDeclaration) {
            // Required to avoid spurious dart errors involving base classes without
            // default constructors.
            this.emit('// @Ignore\n');
            this.fc.visitTypeName(decl.name);
            this.emit('.fakeConstructor$()');
            if (this.extendsClass) {
                // Required to keep the Dart Analyzer happy when a class has subclasses.
                this.emit(': super.fakeConstructor$()');
            }
            this.emit(';\n');
        }
    };
    DeclarationTranspiler.prototype.visitName = function (name) {
        if (base.getEnclosingClass(name) != null) {
            this.visit(name);
            return;
        }
        // Have to rewrite names in this case as we could have conflicts
        // due to needing to support multiple JS modules in a single JS module
        if (name.kind !== ts.SyntaxKind.Identifier) {
            throw 'Internal error: unexpected function name kind:' + name.kind;
        }
        var entry = this.fc.lookupCustomDartTypeName(name);
        if (entry) {
            this.emit(entry.name);
            return;
        }
        this.visit(name);
    };
    DeclarationTranspiler.prototype.notSimpleBagOfProperties = function (type) {
        if (this.tc.getSignaturesOfType(type, ts.SignatureKind.Call).length > 0)
            return true;
        if (this.tc.getSignaturesOfType(type, ts.SignatureKind.Construct).length > 0)
            return true;
        if (type.symbol) {
            var declaration = type.symbol.declarations[0];
            // We have to check the actual declaration as
            if (declaration && declaration.members) {
                var members = declaration.members;
                for (var i = 0; i < members.length; ++i) {
                    var node = members[i];
                    if (base.isStatic(node))
                        return true;
                    switch (node.kind) {
                        case ts.SyntaxKind.PropertyDeclaration:
                        case ts.SyntaxKind.PropertySignature:
                        case ts.SyntaxKind.VariableDeclaration:
                            break;
                        default:
                            return true;
                    }
                }
            }
        }
        return false;
    };
    /**
     * Returns whether all members of the class and all base classes
     */
    DeclarationTranspiler.prototype.hasOnlyProperties = function (decl, outProperties) {
        var type = this.tc.getTypeAtLocation(decl);
        var properties = this.tc.getPropertiesOfType(type);
        var baseTypes = this.tc.getBaseTypes(type);
        if (this.notSimpleBagOfProperties(type))
            return false;
        for (var i = 0; i < baseTypes.length; ++i) {
            var baseType = baseTypes[i];
            if (this.notSimpleBagOfProperties(baseType))
                return false;
        }
        for (var i = 0; i < properties.length; ++i) {
            var symbol = properties[i];
            var node = symbol.valueDeclaration;
            switch (node.kind) {
                case ts.SyntaxKind.PropertyDeclaration:
                case ts.SyntaxKind.PropertySignature:
                case ts.SyntaxKind.VariableDeclaration:
                    var prop = node;
                    if (this.promoteFunctionLikeMembers && isFunctionLikeProperty(prop, this.tc)) {
                        return false;
                    }
                    outProperties.push(prop);
                    break;
                default:
                    return false;
            }
        }
        return outProperties.length > 0;
    };
    DeclarationTranspiler.prototype.visitClassBody = function (decl) {
        var properties = [];
        var isPropertyBag = decl.kind === ts.SyntaxKind.InterfaceDeclaration &&
            this.hasOnlyProperties(decl, properties);
        this.visitMergingOverloads(decl.members);
        if (isPropertyBag) {
            this.emit('external factory');
            this.fc.visitTypeName(decl.name);
            this.emit('({');
            for (var i = 0; i < properties.length; i++) {
                if (i > 0)
                    this.emit(',');
                var p = properties[i];
                this.visit(p.type);
                this.visit(p.name);
            }
            this.emit('});');
        }
    };
    DeclarationTranspiler.prototype.visitMergingOverloads = function (members) {
        var _this = this;
        // TODO(jacobr): merge method overloads.
        var groups = {};
        var orderedGroups = [];
        members.forEach(function (node) {
            var name = '';
            switch (node.kind) {
                case ts.SyntaxKind.Block:
                    // For JS interop we always skip the contents of a block.
                    break;
                case ts.SyntaxKind.PropertyDeclaration:
                case ts.SyntaxKind.PropertySignature:
                case ts.SyntaxKind.VariableDeclaration:
                    {
                        var propertyDecl = node;
                        // We need to emit these as properties not fields.
                        if (!_this.promoteFunctionLikeMembers || !isFunctionLikeProperty(propertyDecl, _this.tc)) {
                            orderedGroups.push([node]);
                            return;
                        }
                        // Convert to a Method.
                        var type = propertyDecl.type;
                        var funcDecl = ts.createNode(ts.SyntaxKind.MethodDeclaration);
                        funcDecl.parent = node.parent;
                        funcDecl.name = propertyDecl.name;
                        switch (type.kind) {
                            case ts.SyntaxKind.FunctionType:
                                var callSignature = type;
                                funcDecl.parameters = callSignature.parameters;
                                funcDecl.type = callSignature.type;
                                // Fall through to the function case using this node
                                node = funcDecl;
                                break;
                            case ts.SyntaxKind.UnionType:
                            case ts.SyntaxKind.TypeLiteral:
                                throw 'Not supported yet';
                            default:
                                throw 'Unexpected case';
                        }
                        name = base.ident(node.name);
                    }
                    break;
                case ts.SyntaxKind.FunctionDeclaration:
                case ts.SyntaxKind.MethodDeclaration:
                case ts.SyntaxKind.MethodSignature:
                case ts.SyntaxKind.FunctionExpression:
                    name = base.ident(node.name);
                    break;
                case ts.SyntaxKind.CallSignature:
                    name = 'call';
                    break;
                case ts.SyntaxKind.Constructor:
                    break;
                case ts.SyntaxKind.ConstructSignature:
                    break;
                case ts.SyntaxKind.IndexSignature:
                    name = '[]';
                    break;
                case ts.SyntaxKind.ClassDeclaration:
                case ts.SyntaxKind.InterfaceDeclaration:
                case ts.SyntaxKind.VariableStatement:
                    orderedGroups.push([node]);
                    return;
                case ts.SyntaxKind.GetAccessor:
                case ts.SyntaxKind.SetAccessor:
                case ts.SyntaxKind.SemicolonClassElement:
                case ts.SyntaxKind.ModuleDeclaration:
                    orderedGroups.push([node]);
                    return;
                default:
                    console.log('Warning: unexpected type... overloads: ' + node.kind + ' ' + node.getText());
                    orderedGroups.push([node]);
                    return;
            }
            var group;
            if (Object.prototype.hasOwnProperty.call(groups, name)) {
                group = groups[name];
            }
            else {
                group = [];
                groups[name] = group;
                orderedGroups.push(group);
            }
            group.push(node);
        });
        orderedGroups.forEach(function (group) {
            if (group.length === 1) {
                _this.visit(group[0]);
                return;
            }
            group.forEach(function (fn) {
                // Emit overrides in a comment that the Dart analyzer can at some point
                // use to improve autocomplete.
                _this.enterCodeComment();
                _this.visit(fn);
                _this.exitCodeComment();
                _this.emit('\n');
            });
            // TODO: actually merge.
            var first = group[0];
            var kind = first.kind;
            var merged = ts.createNode(kind);
            merged.parent = first.parent;
            base.copyLocation(first, merged);
            switch (kind) {
                case ts.SyntaxKind.FunctionDeclaration:
                case ts.SyntaxKind.MethodDeclaration:
                case ts.SyntaxKind.MethodSignature:
                case ts.SyntaxKind.FunctionExpression:
                    var fn = first;
                    merged.name = fn.name;
                    break;
                case ts.SyntaxKind.CallSignature:
                    break;
                case ts.SyntaxKind.Constructor:
                    break;
                case ts.SyntaxKind.ConstructSignature:
                    break;
                case ts.SyntaxKind.IndexSignature:
                    break;
                default:
                    throw 'Unexpected kind:' + kind;
            }
            var mergedParams = first.parameters.map(function (param) { return new merge_1.MergedParameter(param, _this.fc); });
            var mergedType = new merge_1.MergedType(_this.fc);
            mergedType.merge(first.type);
            for (var i = 1; i < group.length; ++i) {
                var signature = group[i];
                mergedType.merge(signature.type);
                var overlap = Math.min(signature.parameters.length, mergedParams.length);
                for (var j = 0; j < overlap; ++j) {
                    mergedParams[j].merge(signature.parameters[j]);
                }
                for (var j = overlap; j < mergedParams.length; ++j) {
                    mergedParams[j].setOptional();
                }
                for (var j = mergedParams.length; j < signature.parameters.length; ++j) {
                    var param = new merge_1.MergedParameter(signature.parameters[j], _this.fc);
                    param.setOptional();
                    mergedParams.push(param);
                }
            }
            merged.parameters = mergedParams.map(function (p) { return p.toParameterDeclaration(); });
            merged.type = mergedType.toTypeNode();
            _this.fc.visit(merged);
        });
    };
    DeclarationTranspiler.prototype.visitNode = function (node) {
        switch (node.kind) {
            case ts.SyntaxKind.ModuleDeclaration:
                var moduleDecl = node;
                this.emit('\n// Module ' + moduleDecl.name.text + '\n');
                this.moduleStack.push(moduleDecl.name.text);
                this.visit(moduleDecl.body);
                this.emit('\n// End module ' + moduleDecl.name.text + '\n');
                this.moduleStack.pop();
                break;
            case ts.SyntaxKind.ExportKeyword:
                // TODO(jacobr): perhaps add a specific Dart annotation to indicate
                // exported members or provide a flag to only generate code for exported
                // members.
                break;
            case ts.SyntaxKind.EnumDeclaration:
                {
                    var decl = node;
                    // The only legal modifier for an enum decl is const.
                    var isConst = decl.modifiers && (decl.modifiers.flags & ts.NodeFlags.Const);
                    if (isConst) {
                        this.reportError(node, 'const enums are not supported');
                    }
                    // In JS interop mode we have to treat enums as JavaScript classes
                    // with static members for each enum constant instead of as first
                    // class enums.
                    this.maybeEmitJsAnnotation(decl);
                    this.emit('class');
                    this.emit(decl.name.text);
                    this.emit('{');
                    var nodes = decl.members;
                    for (var i = 0; i < nodes.length; i++) {
                        this.emit('external static num get');
                        this.visit(nodes[i]);
                        this.emit(';');
                    }
                    this.emit('}');
                }
                break;
            case ts.SyntaxKind.Parameter:
                {
                    var paramDecl = node;
                    if (paramDecl.type && paramDecl.type.kind === ts.SyntaxKind.FunctionType) {
                        // Dart uses "returnType paramName ( parameters )" syntax.
                        var fnType = paramDecl.type;
                        var hasRestParameter = fnType.parameters.some(function (p) { return !!p.dotDotDotToken; });
                        if (!hasRestParameter) {
                            // Dart does not support rest parameters/varargs, degenerate to just "Function".
                            // TODO(jacobr): also consider faking 0 - NUM_FAKE_REST_PARAMETERS
                            // instead.
                            this.visit(fnType.type);
                            this.visit(paramDecl.name);
                            this.visitParameters(fnType.parameters);
                            break;
                        }
                    }
                    if (paramDecl.dotDotDotToken) {
                        // Weak support of varargs that works ok if you have 5 of fewer args.
                        var paramType = void 0;
                        var type = paramDecl.type;
                        if (type) {
                            if (type.kind === ts.SyntaxKind.ArrayType) {
                                var arrayType = type;
                                paramType = arrayType.elementType;
                            }
                            else if (type.kind !== ts.SyntaxKind.AnyKeyword) {
                                throw 'Unexpected type for varargs: ' + type.kind;
                            }
                        }
                        for (var i = 1; i <= DeclarationTranspiler.NUM_FAKE_REST_PARAMETERS; ++i) {
                            if (i > 1) {
                                this.emit(',');
                            }
                            this.visit(paramType);
                            this.emit(base.ident(paramDecl.name) + i);
                        }
                        break;
                    }
                    // TODO(jacobr): should we support
                    if (paramDecl.name.kind === ts.SyntaxKind.ObjectBindingPattern) {
                        this.emit('Object');
                        var pattern = paramDecl.name;
                        var elements = pattern.elements;
                        var name_1 = elements.map(function (e) { return base.ident(e.name); }).join('_');
                        // Warning: this name is unlikely to but could possible overlap with
                        // other parameter names.
                        this.emit(name_1);
                        this.emit('/* ' + pattern.getText() + ' */');
                        break;
                    }
                    if (paramDecl.name.kind !== ts.SyntaxKind.Identifier) {
                        throw 'Unsupported parameter name kind: ' + paramDecl.name.kind;
                    }
                    this.visit(paramDecl.type);
                    this.visit(paramDecl.name);
                }
                break;
            case ts.SyntaxKind.EnumMember:
                {
                    var member = node;
                    this.visit(member.name);
                }
                break;
            case ts.SyntaxKind.ModuleBlock:
                {
                    var block = node;
                    this.visitMergingOverloads(block.statements);
                }
                break;
            case ts.SyntaxKind.VariableDeclarationList:
                {
                    // We have to handle variable declaration lists differently in the case
                    // of JS interop because Dart does not support external variables.
                    var varDeclList = node;
                    this.visitList(varDeclList.declarations, ';');
                }
                break;
            case ts.SyntaxKind.VariableDeclaration:
                {
                    // We have to handle variable declarations differently in the case of JS
                    // interop because Dart does not support external variables.
                    var varDecl = node;
                    this.maybeEmitJsAnnotation(varDecl);
                    this.emit('external');
                    this.visit(varDecl.type);
                    this.emit('get');
                    this.visitName(varDecl.name);
                    if (!this.hasFlag(varDecl.parent, ts.NodeFlags.Const)) {
                        this.emit(';');
                        this.maybeEmitJsAnnotation(varDecl);
                        this.emit('external');
                        this.emit('set');
                        this.visitName(varDecl.name);
                        this.emit('(');
                        this.visit(varDecl.type);
                        this.emit('v)');
                    }
                }
                break;
            case ts.SyntaxKind.StringLiteral:
                {
                    this.emit('String');
                    this.emit('/*');
                    var sLit = node;
                    var text = JSON.stringify(sLit.text);
                    this.emit(text);
                    this.emit('*/');
                }
                break;
            case ts.SyntaxKind.CallSignature:
                {
                    var fn = node;
                    this.emit('external');
                    this.visit(fn.type);
                    this.emit('call');
                    this.visitParameters(fn.parameters);
                    this.emit(';');
                }
                break;
            case ts.SyntaxKind.IndexSignature:
                this.emit('/* Index signature is not yet supported by JavaScript interop.');
                break;
            case ts.SyntaxKind.ExportAssignment:
                // let exportAssignment = <ts.ExportAssignment>node;
                this.emit('/* WARNING: export assignment not yet supported. */\n');
                break;
            case ts.SyntaxKind.TypeAliasDeclaration:
                // Dart does not provide special syntax for definning type alais
                // declarations so we do not emit anything here and resolve alaises
                // to their original types at each usage site.
                break;
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.InterfaceDeclaration:
                {
                    this.extendsClass = false;
                    var classDecl = node;
                    var isInterface = node.kind === ts.SyntaxKind.InterfaceDeclaration;
                    if (isInterface &&
                        base.isFunctionTypedefLikeInterface(classDecl)) {
                        var member = classDecl.members[0];
                        this.visitFunctionTypedefInterface(classDecl.name.text, member, classDecl.typeParameters);
                        break;
                    }
                    var customName = this.fc.lookupCustomDartTypeName(classDecl.name);
                    if (customName && !customName.keep) {
                        this.emit('\n/* Skipping class ' + base.ident(classDecl.name) + '*/\n');
                        break;
                    }
                    this.maybeEmitJsAnnotation(node);
                    if (isInterface ||
                        (classDecl.modifiers && (classDecl.modifiers.flags & ts.NodeFlags.Abstract))) {
                        this.visitClassLike('abstract class', classDecl);
                    }
                    else {
                        this.visitClassLike('class', classDecl);
                    }
                }
                break;
            case ts.SyntaxKind.HeritageClause:
                {
                    var heritageClause = node;
                    if (base.isExtendsClause(heritageClause)) {
                        this.extendsClass = true;
                    }
                    if (base.isExtendsClause(heritageClause)) {
                        this.emit('extends');
                    }
                    else {
                        this.emit('implements');
                    }
                    // Can only have one member for extends clauses.
                    this.visitList(heritageClause.types);
                }
                break;
            case ts.SyntaxKind.ExpressionWithTypeArguments:
                {
                    var exprWithTypeArgs = node;
                    this.visit(exprWithTypeArgs.expression);
                    this.maybeVisitTypeArguments(exprWithTypeArgs);
                }
                break;
            case ts.SyntaxKind.Constructor:
            case ts.SyntaxKind.ConstructSignature:
                {
                    var ctorDecl = node;
                    // Find containing class name.
                    var classDecl = base.getEnclosingClass(ctorDecl);
                    if (!classDecl)
                        this.reportError(ctorDecl, 'cannot find outer class node');
                    this.visitDeclarationMetadata(ctorDecl);
                    this.fc.visitTypeName(classDecl.name);
                    this.visitParameters(ctorDecl.parameters);
                    this.emit(';');
                }
                break;
            case ts.SyntaxKind.PropertyDeclaration:
                this.visitProperty(node);
                break;
            case ts.SyntaxKind.SemicolonClassElement:
                // No-op, don't emit useless declarations.
                break;
            case ts.SyntaxKind.MethodDeclaration:
                this.visitDeclarationMetadata(node);
                this.visitFunctionLike(node);
                break;
            case ts.SyntaxKind.GetAccessor:
                this.visitDeclarationMetadata(node);
                this.visitFunctionLike(node, 'get');
                break;
            case ts.SyntaxKind.SetAccessor:
                this.visitDeclarationMetadata(node);
                this.visitFunctionLike(node, 'set');
                break;
            case ts.SyntaxKind.FunctionDeclaration:
                var funcDecl = node;
                this.visitDeclarationMetadata(funcDecl);
                this.visitFunctionLike(funcDecl);
                break;
            case ts.SyntaxKind.FunctionExpression:
                var funcExpr = node;
                this.visitFunctionLike(funcExpr);
                break;
            case ts.SyntaxKind.PropertySignature:
                var propSig = node;
                this.visitProperty(propSig);
                break;
            case ts.SyntaxKind.MethodSignature:
                var methodSignatureDecl = node;
                this.visitDeclarationMetadata(methodSignatureDecl);
                this.visitFunctionLike(methodSignatureDecl);
                break;
            case ts.SyntaxKind.StaticKeyword:
                // n-op, handled in `visitFunctionLike` and `visitProperty` below.
                break;
            case ts.SyntaxKind.AbstractKeyword:
                // Abstract methods in Dart simply lack implementation,
                // and don't use the 'abstract' modifier
                // Abstract classes are handled in `case ts.SyntaxKind.ClassDeclaration` above.
                break;
            case ts.SyntaxKind.PrivateKeyword:
                // no-op, handled through '_' naming convention in Dart.
                break;
            case ts.SyntaxKind.PublicKeyword:
                // Handled in `visitDeclarationMetadata` below.
                break;
            case ts.SyntaxKind.ProtectedKeyword:
                // Handled in `visitDeclarationMetadata` below.
                break;
            case ts.SyntaxKind.VariableStatement:
                var variableStmt = node;
                this.visit(variableStmt.declarationList);
                this.emit(';');
                break;
            case ts.SyntaxKind.SwitchStatement:
            case ts.SyntaxKind.ArrayLiteralExpression:
            case ts.SyntaxKind.ExpressionStatement:
            case ts.SyntaxKind.EmptyStatement:
                // No need to emit anything for these cases.
                break;
            default:
                return false;
        }
        return true;
    };
    DeclarationTranspiler.prototype.visitFunctionLike = function (fn, accessor) {
        this.fc.pushTypeParameterNames(fn);
        if (base.isStatic(fn)) {
            this.emit('static');
        }
        try {
            this.visit(fn.type);
            if (accessor)
                this.emit(accessor);
            var name_2 = fn.name;
            if (name_2) {
                if (name_2.kind !== ts.SyntaxKind.Identifier) {
                    this.reportError(name_2, 'Unexpected name kind:' + name_2.kind);
                }
                this.fc.visitTypeName(name_2);
            }
            if (fn.typeParameters) {
                this.emit('/*<');
                // Emit the names literally instead of visiting, otherwise they will be replaced with the
                // comment hack themselves.
                this.emit(fn.typeParameters.map(function (p) { return base.ident(p.name); }).join(', '));
                this.emit('>*/');
            }
            // Dart does not even allow the parens of an empty param list on getter
            if (accessor !== 'get') {
                this.visitParameters(fn.parameters);
            }
            else {
                if (fn.parameters && fn.parameters.length > 0) {
                    this.reportError(fn, 'getter should not accept parameters');
                }
            }
            this.emit(';');
        }
        finally {
            this.fc.popTypeParameterNames(fn);
        }
    };
    /**
     * Visit a property declaration.
     * In the special case of property parameters in a constructor, we also allow
     * a parameter to be emitted as a property.
     * We have to emit properties as getter setter pairs as Dart does not support
     * external fields.
     * In the special case of property parameters in a constructor, we also allow a parameter to be
     * emitted as a property.
     */
    DeclarationTranspiler.prototype.visitProperty = function (decl, isParameter) {
        if (isParameter === void 0) { isParameter = false; }
        var isStatic = base.isStatic(decl);
        this.emit('external');
        if (isStatic)
            this.emit('static');
        this.visit(decl.type);
        this.emit('get');
        this.visitName(decl.name);
        this.emit(';');
        this.emit('external');
        if (isStatic)
            this.emit('static');
        this.emit('set');
        this.visitName(decl.name);
        this.emit('(');
        this.visit(decl.type);
        this.emit('v');
        this.emit(')');
        this.emit(';');
    };
    DeclarationTranspiler.prototype.visitClassLike = function (keyword, decl) {
        var _this = this;
        this.emit(keyword);
        this.fc.visitTypeName(decl.name);
        if (decl.typeParameters) {
            this.emit('<');
            this.visitList(decl.typeParameters);
            this.emit('>');
        }
        this.visitEachIfPresent(decl.heritageClauses);
        this.emit('{');
        this.maybeEmitFakeConstructors(decl);
        // Synthesize explicit properties for ctor with 'property parameters'
        var synthesizePropertyParam = function (param) {
            if (_this.hasFlag(param.modifiers, ts.NodeFlags.Public) ||
                _this.hasFlag(param.modifiers, ts.NodeFlags.Private) ||
                _this.hasFlag(param.modifiers, ts.NodeFlags.Protected)) {
                // TODO: we should enforce the underscore prefix on privates
                _this.visitProperty(param, true);
            }
        };
        decl.members
            .filter(base.isConstructor)
            .forEach(function (ctor) {
            return ctor.parameters.forEach(synthesizePropertyParam);
        });
        this.visitClassBody(decl);
        this.emit('}');
    };
    DeclarationTranspiler.prototype.visitDeclarationMetadata = function (decl) {
        this.visitEachIfPresent(decl.modifiers);
        switch (decl.kind) {
            case ts.SyntaxKind.Constructor:
            case ts.SyntaxKind.ConstructSignature:
                this.emit('external factory');
                break;
            case ts.SyntaxKind.ArrowFunction:
            case ts.SyntaxKind.CallSignature:
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.SetAccessor:
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.MethodSignature:
            case ts.SyntaxKind.PropertySignature:
            case ts.SyntaxKind.FunctionDeclaration:
                if (!base.getEnclosingClass(decl)) {
                    this.maybeEmitJsAnnotation(decl);
                }
                this.emit('external');
                break;
            default:
                throw 'Unexpected declaration kind:' + decl.kind;
        }
    };
    /**
     * Handles a function typedef-like interface, i.e. an interface that only declares a single
     * call signature, by translating to a Dart `typedef`.
     */
    DeclarationTranspiler.prototype.visitFunctionTypedefInterface = function (name, signature, typeParameters) {
        this.emit('typedef');
        if (signature.type) {
            this.visit(signature.type);
        }
        this.emit(name);
        if (typeParameters) {
            this.emit('<');
            this.visitList(typeParameters);
            this.emit('>');
        }
        this.visitParameters(signature.parameters);
        this.emit(';');
    };
    DeclarationTranspiler.NUM_FAKE_REST_PARAMETERS = 5;
    return DeclarationTranspiler;
}(base.TranspilerBase));
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = DeclarationTranspiler;

//# sourceMappingURL=declaration.js.map
