"use strict";
var fs = require('fs');
var path = require('path');
var ts = require('typescript');
var mkdirp_1 = require('./mkdirp');
var declaration_1 = require('./declaration');
var merge = require('./merge');
var module_1 = require('./module');
var type_1 = require('./type');
var facade_converter_1 = require('./facade_converter');
var dartStyle = require('dart-style');
exports.COMPILER_OPTIONS = {
    allowNonTsExtensions: true,
    experimentalDecorators: true,
    module: ts.ModuleKind.CommonJS,
    target: ts.ScriptTarget.ES5,
};
/**
 * Context to ouput code into.
 */
(function (OutputContext) {
    OutputContext[OutputContext["Import"] = 0] = "Import";
    OutputContext[OutputContext["Header"] = 1] = "Header";
    OutputContext[OutputContext["Default"] = 2] = "Default";
})(exports.OutputContext || (exports.OutputContext = {}));
var OutputContext = exports.OutputContext;
var NUM_OUTPUT_CONTEXTS = 3;
var Transpiler = (function () {
    function Transpiler(options) {
        if (options === void 0) { options = {}; }
        this.options = options;
        // Comments attach to all following AST nodes before the next 'physical' token. Track the earliest
        // offset to avoid printing comments multiple times.
        this.lastCommentIdx = -1;
        this.errors = [];
        this.nameRewriter = new facade_converter_1.NameRewriter();
        this.fc = new facade_converter_1.FacadeConverter(this, options.typingsRoot, this.nameRewriter, options.useHtml);
        this.declarationTranspiler = new declaration_1.default(this, this.fc, options.enforceUnderscoreConventions, options.promoteFunctionLikeMembers);
        this.transpilers = [
            this.declarationTranspiler,
            new module_1.default(this, this.fc, options.generateLibraryName),
            new type_1.default(this, this.fc),
        ];
    }
    /**
     * Transpiles the given files to Dart.
     * @param fileNames The input files.
     * @param destination Location to write files to. Creates files next to their sources if absent.
     */
    Transpiler.prototype.transpile = function (fileNames, destination) {
        var _this = this;
        if (this.options.basePath) {
            this.options.basePath = this.normalizeSlashes(path.resolve(this.options.basePath));
        }
        fileNames = fileNames.map(function (f) { return _this.normalizeSlashes(f); });
        var host = this.createCompilerHost();
        if (this.options.basePath && destination === undefined) {
            throw new Error('Must have a destination path when a basePath is specified ' + this.options.basePath);
        }
        var destinationRoot = destination || this.options.basePath || '';
        var program = ts.createProgram(fileNames, this.getCompilerOptions(), host);
        this.fc.setTypeChecker(program.getTypeChecker());
        this.declarationTranspiler.setTypeChecker(program.getTypeChecker());
        // Only write files that were explicitly passed in.
        var fileSet = {};
        fileNames.forEach(function (f) { return fileSet[f] = true; });
        var sourceFiles = program.getSourceFiles().filter(function (sourceFile) { return fileSet[sourceFile.fileName]; });
        this.errors = [];
        sourceFiles.forEach(function (f) {
            var dartCode = _this.translate(f);
            var outputFile = _this.getOutputPath(path.resolve(f.fileName), destinationRoot);
            mkdirp_1.default(path.dirname(outputFile));
            fs.writeFileSync(outputFile, dartCode);
        });
        this.checkForErrors(program);
    };
    Transpiler.prototype.translateProgram = function (program) {
        var _this = this;
        this.fc.setTypeChecker(program.getTypeChecker());
        this.declarationTranspiler.setTypeChecker(program.getTypeChecker());
        var paths = {};
        this.errors = [];
        program.getSourceFiles()
            .filter(function (sourceFile) {
            return (!sourceFile.fileName.match(/\.d\.ts$/) && !!sourceFile.fileName.match(/\.[jt]s$/));
        })
            .forEach(function (f) { return paths[f.fileName] = _this.translate(f); });
        this.checkForErrors(program);
        return paths;
    };
    Transpiler.prototype.getCompilerOptions = function () {
        var opts = {};
        for (var _i = 0, _a = Object.keys(exports.COMPILER_OPTIONS); _i < _a.length; _i++) {
            var k = _a[_i];
            opts[k] = exports.COMPILER_OPTIONS[k];
        }
        opts.rootDir = this.options.basePath;
        return opts;
    };
    Transpiler.prototype.createCompilerHost = function () {
        var defaultLibFileName = ts.getDefaultLibFileName(exports.COMPILER_OPTIONS);
        defaultLibFileName = this.normalizeSlashes(defaultLibFileName);
        var compilerHost = {
            getSourceFile: function (sourceName, languageVersion) {
                var sourcePath = sourceName;
                if (sourceName === defaultLibFileName) {
                    sourcePath = ts.getDefaultLibFilePath(exports.COMPILER_OPTIONS);
                }
                if (!fs.existsSync(sourcePath))
                    return undefined;
                var contents = fs.readFileSync(sourcePath, 'UTF-8');
                return ts.createSourceFile(sourceName, contents, exports.COMPILER_OPTIONS.target, true);
            },
            writeFile: function (name, text, writeByteOrderMark) { fs.writeFile(name, text); },
            fileExists: function (filename) { return fs.existsSync(filename); },
            readFile: function (filename) { return fs.readFileSync(filename, 'utf-8'); },
            getDefaultLibFileName: function () { return defaultLibFileName; },
            useCaseSensitiveFileNames: function () { return true; },
            getCanonicalFileName: function (filename) { return filename; },
            getCurrentDirectory: function () { return ''; },
            getNewLine: function () { return '\n'; },
        };
        compilerHost.resolveModuleNames = getModuleResolver(compilerHost);
        return compilerHost;
    };
    // Visible for testing.
    Transpiler.prototype.getOutputPath = function (filePath, destinationRoot) {
        var relative = this.getRelativeFileName(filePath);
        var dartFile = relative.replace(/.(js|es6|d\.ts|ts)$/, '.dart');
        return this.normalizeSlashes(path.join(destinationRoot, dartFile));
    };
    Transpiler.prototype.pushContext = function (context) { this.outputStack.push(this.outputs[context]); };
    Transpiler.prototype.popContext = function () {
        if (this.outputStack.length === 0) {
            this.reportError(null, 'Attempting to pop output stack when already empty');
        }
        this.outputStack.pop();
    };
    Transpiler.prototype.translate = function (sourceFile) {
        this.currentFile = sourceFile;
        this.outputs = [];
        this.outputStack = [];
        this.importsEmitted = {};
        for (var i = 0; i < NUM_OUTPUT_CONTEXTS; ++i) {
            this.outputs.push(new Output());
        }
        this.lastCommentIdx = -1;
        merge.normalizeSourceFile(sourceFile);
        this.pushContext(OutputContext.Default);
        this.visit(sourceFile);
        this.popContext();
        if (this.outputStack.length > 0) {
            this.reportError(sourceFile, 'Internal error managing output contexts. ' +
                'Inconsistent push and pop context calls.');
        }
        var result = '';
        for (var _i = 0, _a = this.outputs; _i < _a.length; _i++) {
            var output = _a[_i];
            result += output.getResult();
        }
        return this.formatCode(result, sourceFile);
    };
    Transpiler.prototype.formatCode = function (code, context) {
        var result = dartStyle.formatCode(code);
        if (result.error) {
            this.reportError(context, result.error);
        }
        return result.code;
    };
    Transpiler.prototype.checkForErrors = function (program) {
        var _this = this;
        var errors = this.errors;
        var diagnostics = program.getGlobalDiagnostics().concat(program.getSyntacticDiagnostics());
        if ((errors.length || diagnostics.length)) {
            // Only report semantic diagnostics if facade generation failed; this
            // code is not a generic compiler, so only yields TS errors if they could
            // be the cause of facade generation issues.
            // This greatly speeds up tests and execution.
            diagnostics = diagnostics.concat(program.getSemanticDiagnostics());
        }
        var diagnosticErrs = diagnostics.map(function (d) {
            var msg = '';
            if (d.file) {
                var pos = d.file.getLineAndCharacterOfPosition(d.start);
                var fn = _this.getRelativeFileName(d.file.fileName);
                msg += " " + fn + ":" + (pos.line + 1) + ":" + (pos.character + 1);
            }
            msg += ': ';
            msg += ts.flattenDiagnosticMessageText(d.messageText, '\n');
            return msg;
        });
        if (diagnosticErrs.length)
            errors = errors.concat(diagnosticErrs);
        if (errors.length) {
            var e = new Error(errors.join('\n'));
            e.name = 'DartFacadeError';
            throw e;
        }
    };
    /**
     * Returns `filePath`, relativized to the program's `basePath`.
     * @param filePath Optional path to relativize, defaults to the current file's path.
     */
    Transpiler.prototype.getRelativeFileName = function (filePath) {
        if (filePath === undefined)
            filePath = path.resolve(this.currentFile.fileName);
        // TODO(martinprobst): Use path.isAbsolute on node v0.12.
        if (this.normalizeSlashes(path.resolve('/x/', filePath)) !== filePath) {
            return filePath; // already relative.
        }
        var base = this.options.basePath || '';
        if (filePath.indexOf(base) !== 0 && !filePath.match(/\.d\.ts$/)) {
            throw new Error("Files must be located under base, got " + filePath + " vs " + base);
        }
        return this.normalizeSlashes(path.relative(base, filePath));
    };
    Object.defineProperty(Transpiler.prototype, "currentOutput", {
        get: function () { return this.outputStack[this.outputStack.length - 1]; },
        enumerable: true,
        configurable: true
    });
    Transpiler.prototype.emit = function (s) { this.currentOutput.emit(s); };
    Transpiler.prototype.emitNoSpace = function (s) { this.currentOutput.emitNoSpace(s); };
    Transpiler.prototype.enterCodeComment = function () { return this.currentOutput.enterCodeComment(); };
    Transpiler.prototype.exitCodeComment = function () { return this.currentOutput.exitCodeComment(); };
    Transpiler.prototype.reportError = function (n, message) {
        var file = n.getSourceFile() || this.currentFile;
        var fileName = this.getRelativeFileName(file.fileName);
        var start = n.getStart(file);
        var pos = file.getLineAndCharacterOfPosition(start);
        // Line and character are 0-based.
        var fullMessage = fileName + ":" + (pos.line + 1) + ":" + (pos.character + 1) + ": " + message;
        if (this.options.failFast)
            throw new Error(fullMessage);
        this.errors.push(fullMessage);
    };
    Transpiler.prototype.visit = function (node) {
        var _this = this;
        if (!node)
            return;
        var comments = ts.getLeadingCommentRanges(this.currentFile.text, node.getFullStart());
        if (comments) {
            comments.forEach(function (c) {
                // Warning: the following check means that comments will only be
                // emitted correctly if Dart code is emitted in the same order it
                // appeared in the JavaScript AST.
                if (c.pos <= _this.lastCommentIdx)
                    return;
                _this.lastCommentIdx = c.pos;
                var text = _this.currentFile.text.substring(c.pos, c.end);
                _this.currentOutput.maybeLineBreak();
                _this.emit(_this.translateComment(text));
                if (c.hasTrailingNewLine)
                    _this.emitNoSpace('\n');
            });
        }
        for (var i = 0; i < this.transpilers.length; i++) {
            if (this.transpilers[i].visitNode(node))
                return;
        }
        this.reportError(node, 'Unsupported node type ' + ts.SyntaxKind[node.kind] + ': ' + node.getFullText());
    };
    Transpiler.prototype.normalizeSlashes = function (path) { return path.replace(/\\/g, '/'); };
    Transpiler.prototype.translateComment = function (comment) {
        comment = comment.replace(/\{@link ([^\}]+)\}/g, '[$1]');
        // Remove the following tags and following comments till end of line.
        comment = comment.replace(/@param.*$/gm, '');
        comment = comment.replace(/@throws.*$/gm, '');
        comment = comment.replace(/@return.*$/gm, '');
        // Remove the following tags.
        comment = comment.replace(/@module/g, '');
        comment = comment.replace(/@description/g, '');
        comment = comment.replace(/@deprecated/g, '');
        // Hack to make comment indentation does not look as bad when dart format is
        // run on files with multiple nested modules.
        comment = comment.replace(/\n  ( +[*])/g, '\n$1');
        return comment;
    };
    return Transpiler;
}());
exports.Transpiler = Transpiler;
function getModuleResolver(compilerHost) {
    return function (moduleNames, containingFile) {
        var res = [];
        for (var _i = 0, moduleNames_1 = moduleNames; _i < moduleNames_1.length; _i++) {
            var mod = moduleNames_1[_i];
            var lookupRes = ts.nodeModuleNameResolver(mod, containingFile, exports.COMPILER_OPTIONS, compilerHost);
            if (lookupRes.resolvedModule) {
                res.push(lookupRes.resolvedModule);
                continue;
            }
            lookupRes = ts.classicNameResolver(mod, containingFile, exports.COMPILER_OPTIONS, compilerHost);
            if (lookupRes.resolvedModule) {
                res.push(lookupRes.resolvedModule);
                continue;
            }
            res.push(undefined);
        }
        return res;
    };
}
exports.getModuleResolver = getModuleResolver;
var Output = (function () {
    function Output() {
        this.result = '';
        this.firstColumn = true;
        this.insideCodeComment = false;
        this.codeCommentResult = '';
    }
    /**
     * Line break if the current line is not empty.
     */
    Output.prototype.maybeLineBreak = function () {
        if (this.insideCodeComment) {
            // Avoid line breaks inside code comments.
            return;
        }
        if (!this.firstColumn) {
            this.emitNoSpace('\n');
        }
    };
    Output.prototype.emit = function (str) {
        this.emitNoSpace(' ');
        this.emitNoSpace(str);
    };
    Output.prototype.emitNoSpace = function (str) {
        if (str.length === 0)
            return;
        if (this.insideCodeComment) {
            this.codeCommentResult += str;
            return;
        }
        this.result += str;
        this.firstColumn = str[str.length - 1] === '\n';
    };
    Output.prototype.enterCodeComment = function () {
        if (this.insideCodeComment) {
            throw 'Cannot nest code comments';
        }
        this.insideCodeComment = true;
        this.codeCommentResult = '';
    };
    Output.prototype.exitCodeComment = function () {
        if (!this.insideCodeComment) {
            throw 'Exit code comment called while not within a code comment.';
        }
        this.insideCodeComment = false;
        this.emit('/*');
        var result = dartStyle.formatCode(this.codeCommentResult);
        if (result.error) {
            // Fall back to the raw expression if the formatter returns an error.
            // Ideally the formatter would handle a wider range of dart expressions.
            this.emit(this.codeCommentResult.trim());
        }
        else {
            this.emit(result.code.trim());
        }
        this.emit('*/');
        // Don't really need an exact column, just need to track
        // that we aren't on the first column.
        this.firstColumn = false;
        this.codeCommentResult = '';
    };
    Output.prototype.getResult = function () {
        if (this.insideCodeComment) {
            throw 'Code comment not property terminated.';
        }
        return this.result;
    };
    return Output;
}());
// CLI entry point
if (require.main === module) {
    var args = require('minimist')(process.argv.slice(2), { base: 'string' });
    try {
        var transpiler = new Transpiler(args);
        console.error('Transpiling', args._, 'to', args.destination);
        transpiler.transpile(args._, args.destination);
    }
    catch (e) {
        if (e.name !== 'DartFacadeError')
            throw e;
        console.error(e.message);
        process.exit(1);
    }
}

//# sourceMappingURL=main.js.map
