"use strict";
var ts = require('typescript');
var base = require('./base');
/**
 * To support arbitrary d.ts files in Dart we often have to merge two TypeScript
 * types into a single Dart type because Dart lacks features such as method
 * overloads, type aliases, and union types.
 */
var MergedType = (function () {
    function MergedType(fc) {
        this.fc = fc;
        this.types = {};
    }
    MergedType.prototype.merge = function (t) {
        if (t) {
            // TODO(jacobr): get a better unique name for a type.
            switch (t.kind) {
                case ts.SyntaxKind.UnionType:
                    var union = t;
                    union.types.forEach(this.merge.bind(this));
                    return;
                case ts.SyntaxKind.LastTypeNode:
                    this.merge(t.type);
                    return;
                case ts.SyntaxKind.TypePredicate:
                    this.merge(t.type);
                    return;
                case ts.SyntaxKind.TypeReference:
                    // We need to follow Alais types as Dart does not support them for non
                    // function types. TODO(jacobr): handle them for Function types?
                    var typeRef = t;
                    var decl = this.fc.getDeclaration(typeRef.typeName);
                    if (decl !== null && decl.kind === ts.SyntaxKind.TypeAliasDeclaration) {
                        var alias = decl;
                        if (typeRef.typeArguments) {
                            throw 'TypeReference with arguements not supported yet:' + t.getText();
                        }
                        this.merge(alias.type);
                        return;
                    }
                    break;
                default:
                    break;
            }
            this.types[this.fc.generateDartTypeName(t)] = t;
        }
    };
    MergedType.prototype.toTypeNode = function () {
        var names = Object.getOwnPropertyNames(this.types);
        if (names.length === 0) {
            return null;
        }
        if (names.length === 1) {
            return this.types[names[0]];
        }
        var union = ts.createNode(ts.SyntaxKind.UnionType);
        base.copyLocation(this.types[names[0]], union);
        union.types = [];
        for (var i = 0; i < names.length; ++i) {
            union.types.push(this.types[names[i]]);
        }
        return union;
    };
    return MergedType;
}());
exports.MergedType = MergedType;
/**
 * Handle a parameter that is the result of merging parameter declarations from
 * multiple method overloads.
 */
var MergedParameter = (function () {
    function MergedParameter(param, fc) {
        this.name = {};
        this.optional = false;
        this.type = new MergedType(fc);
        this.textRange = param;
        this.merge(param);
    }
    MergedParameter.prototype.merge = function (param) {
        this.name[base.ident(param.name)] = true;
        if (!this.optional) {
            this.optional = !!param.questionToken;
        }
        this.type.merge(param.type);
    };
    MergedParameter.prototype.toParameterDeclaration = function () {
        var ret = ts.createNode(ts.SyntaxKind.Parameter);
        var nameIdentifier = ts.createNode(ts.SyntaxKind.Identifier);
        nameIdentifier.text = Object.getOwnPropertyNames(this.name).join('_');
        ret.name = nameIdentifier;
        if (this.optional) {
            ret.questionToken = ts.createNode(ts.SyntaxKind.QuestionToken);
        }
        base.copyLocation(this.textRange, ret);
        ret.type = this.type.toTypeNode();
        return ret;
    };
    MergedParameter.prototype.setOptional = function () { this.optional = true; };
    return MergedParameter;
}());
exports.MergedParameter = MergedParameter;
/**
 * Normalize a SourceFile
 */
function normalizeSourceFile(f) {
    var modules = {};
    // Merge top level modules.
    for (var i = 0; i < f.statements.length; ++i) {
        var statement = f.statements[i];
        if (statement.kind !== ts.SyntaxKind.ModuleDeclaration)
            continue;
        var moduleDecl = statement;
        var name_1 = moduleDecl.name.text;
        if (modules.hasOwnProperty(name_1)) {
            var srcBody = modules[name_1].body;
            var srcBodyBlock = void 0;
            this.assert(srcBody, srcBody.kind === ts.SyntaxKind.ModuleBlock, 'Module body must be a module block.');
            srcBodyBlock = srcBody;
            var body = moduleDecl.body;
            if (body.kind === ts.SyntaxKind.ModuleBlock) {
                var bodyBlock = body;
                Array.prototype.push.apply(srcBodyBlock.statements, bodyBlock.statements);
            }
            else {
                // moduleDecl.body is a ModuleDeclaration.
                srcBodyBlock.statements.push(moduleDecl.body);
            }
            f.statements.splice(i, 1);
            i--;
        }
        else {
            modules[name_1] = moduleDecl;
        }
    }
    function addModifier(n, modifier) {
        if (!n.modifiers) {
            n.modifiers = [];
            n.modifiers.flags = 0;
        }
        modifier.parent = n;
        n.modifiers.push(modifier);
    }
    function mergeVariablesIntoClasses(n, classes) {
        switch (n.kind) {
            case ts.SyntaxKind.VariableStatement:
                var statement = n;
                statement.declarationList.declarations.forEach(function (declaration) {
                    if (declaration.name.kind === ts.SyntaxKind.Identifier) {
                        var name_2 = (declaration.name).text;
                        if (classes.hasOwnProperty(name_2)) {
                            var existing_1 = classes[name_2];
                            var members_1 = existing_1.members;
                            if (declaration.type) {
                                var type = declaration.type;
                                if (type.kind === ts.SyntaxKind.TypeLiteral) {
                                    removeNode(n);
                                    var literal = type;
                                    literal.members.forEach(function (member) {
                                        switch (member.kind) {
                                            case ts.SyntaxKind.ConstructSignature:
                                                var signature = member;
                                                var constructor = ts.createNode(ts.SyntaxKind.Constructor);
                                                constructor.parameters = signature.parameters;
                                                constructor.type = signature.type;
                                                base.copyLocation(signature, constructor);
                                                constructor.typeParameters = signature.typeParameters;
                                                constructor.parent = existing_1;
                                                members_1.push(constructor);
                                                break;
                                            case ts.SyntaxKind.Constructor:
                                                member.parent = existing_1.parent;
                                                members_1.push(member);
                                                break;
                                            case ts.SyntaxKind.MethodSignature:
                                                member.parent = existing_1.parent;
                                                members_1.push(member);
                                                break;
                                            case ts.SyntaxKind.PropertySignature:
                                                addModifier(member, ts.createNode(ts.SyntaxKind.StaticKeyword));
                                                member.parent = existing_1;
                                                members_1.push(member);
                                                break;
                                            case ts.SyntaxKind.IndexSignature:
                                                member.parent = existing_1.parent;
                                                members_1.push(member);
                                                break;
                                            case ts.SyntaxKind.CallSignature:
                                                member.parent = existing_1.parent;
                                                members_1.push(member);
                                                break;
                                            default:
                                                throw 'Unhandled TypeLiteral member type:' + member.kind;
                                        }
                                    });
                                }
                            }
                        }
                    }
                    else {
                        throw 'Unexpected VariableStatement identifier kind';
                    }
                });
                break;
            case ts.SyntaxKind.ModuleBlock:
                ts.forEachChild(n, function (child) { return mergeVariablesIntoClasses(child, classes); });
                break;
            default:
                break;
        }
    }
    function removeFromArray(nodes, v) {
        for (var i = 0, len = nodes.length; i < len; ++i) {
            if (nodes[i] === v) {
                nodes.splice(i, 1);
                break;
            }
        }
    }
    function removeNode(n) {
        var parent = n.parent;
        switch (parent.kind) {
            case ts.SyntaxKind.ModuleBlock:
                var block = parent;
                removeFromArray(block.statements, n);
                break;
            case ts.SyntaxKind.SourceFile:
                var sourceFile = parent;
                removeFromArray(sourceFile.statements, n);
                break;
            default:
                throw 'removeNode not implemented for kind:' + parent.kind;
        }
    }
    function makeCallableClassesImplementFunction(decl) {
        if (base.isCallable(decl)) {
            // Modify the AST to explicitly state that the class implements Function
            if (!decl.heritageClauses) {
                decl.heritageClauses = [];
                base.copyLocation(decl, decl.heritageClauses);
            }
            var clauses = decl.heritageClauses;
            var clause = base.arrayFindPolyfill(clauses, function (c) { return c.token !== ts.SyntaxKind.ExtendsKeyword ||
                decl.kind === ts.SyntaxKind.InterfaceDeclaration; });
            if (clause == null) {
                clause = ts.createNode(ts.SyntaxKind.HeritageClause);
                clause.token = decl.kind === ts.SyntaxKind.InterfaceDeclaration ?
                    ts.SyntaxKind.ExtendsKeyword :
                    ts.SyntaxKind.ImplementsKeyword;
                clause.types = [];
                clause.parent = decl;
                base.copyLocation(decl, clause);
                clauses.push(clause);
            }
            var functionType = ts.createNode(ts.SyntaxKind.ExpressionWithTypeArguments);
            functionType.parent = clause;
            base.copyLocation(clause, functionType);
            var fn = ts.createNode(ts.SyntaxKind.Identifier);
            fn.text = 'Function';
            fn.parent = functionType;
            base.copyLocation(functionType, fn);
            functionType.expression = fn;
            clause.types.push(functionType);
        }
    }
    function gatherClasses(n, classes) {
        switch (n.kind) {
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.InterfaceDeclaration:
                var classDecl = n;
                var name_3 = classDecl.name.text;
                // TODO(jacobr): validate that the classes have consistent
                // modifiers, etc.
                if (classes.hasOwnProperty(name_3)) {
                    var existing_2 = classes[name_3];
                    classDecl.members.forEach(function (e) {
                        existing_2.members.push(e);
                        e.parent = existing_2;
                    });
                    removeNode(classDecl);
                }
                else {
                    classes[name_3] = classDecl;
                    // Perform other class level post processing here.
                    makeCallableClassesImplementFunction(classDecl);
                }
                break;
            case ts.SyntaxKind.ModuleDeclaration:
            case ts.SyntaxKind.SourceFile:
                var moduleClasses_1 = {};
                ts.forEachChild(n, function (child) { return gatherClasses(child, moduleClasses_1); });
                ts.forEachChild(n, function (child) { return mergeVariablesIntoClasses(child, moduleClasses_1); });
                break;
            case ts.SyntaxKind.ModuleBlock:
                ts.forEachChild(n, function (child) { return gatherClasses(child, classes); });
                break;
            default:
                break;
        }
    }
    gatherClasses(f, {});
}
exports.normalizeSourceFile = normalizeSourceFile;

//# sourceMappingURL=merge.js.map
